from pedal.questions.graders import FunctionGrader
from pedal.questions.questions import Question

def grade_function_question(function_name, signature, tests, instructions=None, config=None,
                            max_points=None, definition_points=None, components_points=None,
                            max_components_points=None, unit_test_type_points=None,
                            unit_test_value_points=None, unit_test_total_points=None,
                            unit_test_type_ratio=None, unit_test_completion_points=None
                            ):
    """

    Args:
        function_name:
        signature:
        tests:
        config:

    Returns:

    """
    grader = FunctionGrader(function_name, signature, tests, config,
                            max_points, definition_points, components_points,
                            max_components_points, unit_test_type_points,
                            unit_test_value_points, unit_test_total_points,
                            unit_test_type_ratio, unit_test_completion_points)
    question = Question(function_name, instructions, grader)
    return grader._test(question)


import os
import sys

from pedal.utilities.out_files import parse_out_file, generate_out_file


def _make_final(environment):
    """ Determine the final environment's section """
    if environment is None:
        return 'final'
    else:
        return f"{environment}.final"


class ReportVerifier:
    """ Helper class for reading and processing a '.out' file. """
    def __init__(self, path, environment):
        self.outfile = parse_out_file(path)
        self.environment = environment

    def get_final(self):
        """ Get the expected final feedback. """
        final_environment = _make_final(self.environment)
        if final_environment in self.outfile:
            return self.outfile[final_environment].items()
        elif 'final' in self.outfile:
            return self.outfile['final'].items()
        return {}


def generate_report_out(path, environment, report):
    result = {}
    # Allow graceful appending
    if path is not None and os.path.exists(path):
        result = parse_out_file(path)
    final_environment = _make_final(environment)
    try:
        fields = report.result.to_json()
    except Exception as e:
        raise report
    # TODO: Hack, just remove the data field for simplicity - fix this later
    fields.pop('data')
    fields.pop('hide_correctness')
    result[final_environment] = fields
    # TODO: Support ``forbid`` and ``available`` fields for finer grained
    #   output checking.
    if path is not None:
        with open(path, 'w') as outfile:
            generate_out_file(outfile, result)
    else:
        generate_out_file(sys.stdout, result)
        return result



from pedal.command_line.modes import AbstractPipeline, Bundle
from pedal.core.config_job import JobConfig

class SimplePipeline(AbstractPipeline):
    def process_output(self):
        return self

def run_job(**config) -> list[Bundle]:
    """
    Runs a Pedal job with the given configuration.
    """
    pipeline = SimplePipeline(JobConfig(mode="simple", **config))
    pipeline.execute()
    return pipeline.submissions



"""
Configuration support for Pedal!

Pedal users should be able to specify configuration in all the following ways:
- Via a configuration file (pedal.config, or CLI parameter)
- Via environment variables
- Via a cli args
- Via keyword arguments to the main Pedal entry point functions
- Via a configuration object

We can use argparse.parse_known_args() to parse the command line arguments.

Pedal is used to execute a Job, which has one or more Bundles.
Jobs are executed through one of the Pipelines, which are different modes.
Jobs are configured via JobConfig, either through the CLI or a config file.

Bundle executions occurs in an Environment, with a submission and ICS.
The result will be a BundleResult, which has data, output, errors, and resolution.

An EnvironmentConfig determines how a Bundle is executed.
The Job sets up the EnvironmentConfig, to dispatch to the setup_environment.
"""

import argparse
from collections import defaultdict
from typing import Optional, List, Dict, Any, Type, TypeVar
from pedal.command_line.modes import MODES
from dataclasses import dataclass, field

from pedal.environments import ALL_ENVIRONMENTS
from pedal.utilities.out_files import parse_out_file


def metadata(help: str, choices: List[str] = None,
             optional: bool = True, short_form: str = None,
             nargs: str = None, action: str = 'store'):
    # Build up a typesafe dictionary from the arguments
    result = {
        "help": help,
        "optional": optional,
        "short_form": short_form,
        "action": action
    }
    if action not in ('store_true', 'store_false'):
        result['choices'] = choices
        result["nargs"] = nargs
    return result


T = TypeVar("T", bound="EnvironmentConfig")

MAIN_PARSER_DESCRIPTION = "Run instructor control script on student submission."

@dataclass
class JobConfig:
    mode: MODES = field(
        metadata=metadata(
            optional=False,
            help="What kind of Pedal analysis you're running. See the description of each mode above.",
            choices=list(MODES.PIPELINES)
        )
    )
    # This field is only required if we're in reduced mode.
    instructor: Optional[str] = field(
        metadata=metadata(
            optional=False,
            help='The path to the instructor control script, or multiple scripts.'
        )
    )
    submissions: str = field(
        default="submissions",
        metadata=metadata(
            optional=False,
            help='The path to the student submissions.'
                        ' Defaults to a folder named '
                        'submissions adjacent to the '
                        'instructor control script.',
            nargs="?"
        )
    )
    alternate_filenames: str = field(
        default="",
        metadata=metadata(
            help="A semicolon separated list of potential filenames to try if the main isn't found.",
        )
    )
    config: Optional[str] = field(
        default=None,
        metadata=metadata(
            help="Uses the configuration file to get settings."
        )
    )
    output: str = field(
        default="stdout",
        metadata=metadata(
            help='The output file path for the result of the instructor script (not the student executions output). Defaults to stdout.',
            short_form="o"
        )
    )
    create_output: bool = field(
        default=False,
        metadata=metadata(
            help="In verify mode, creates any missing outputs.",
            short_form="m",
            action="store_true"
        )
    )
    environment: str = field(
        default="standard",
        metadata=metadata(
            help="Sets the environment context for this script, which"
                 " can run special setups and override tools as"
                 " needed.",
            choices=ALL_ENVIRONMENTS
        )
    )
    instructor_name: str = field(
        default=None,
        metadata=metadata(
            help="Sets the name of the instructor file to something"
                 " more friendly. If not given, then will default"
                 " to the instructor filename.",
        )
    )
    progsnap_profile: str = field(
        default="blockpy", # TODO: Change this biased default
        metadata=metadata(
            help="Uses the given profile's default settings for"
                 " loading in a ProgSnap2 dataset",
        )
    )
    include_scripts: Optional[str] = field(
        default=None,
        metadata=metadata(
            help="An optional filter to only include certain scripts",
        )
    )
    limit: Optional[int] = field(
        default=None,
        metadata=metadata(
            help="An optional limit to how many submissions are run. Mostly for testing purposes.",
        )
    )
    resolver: str = field(
        default="resolve",
        metadata=metadata(
            help="Choose a different resolver to use (the name of a function defined in the instructor control script).",
        )
    )
    submission_direct: bool = field(
        default=False,
        metadata=metadata(
            help="Give the student code directly instead of loading from a file.",
            action="store_true"
        )
    )
    instructor_direct: bool = field(
        default=False,
        metadata=metadata(
            help="Give the instructor code directly instead of loading from a file.",
            action="store_true"
        )
    )
    skip_tifa: bool = field(
        default=False,
        metadata=metadata(
            help="Skip using TIFA in the environment",
            action="store_true"
        )
    )
    skip_run: bool = field(
        default=False,
        metadata=metadata(
            help="Skip automatically running student code in the environment",
            action="store_true"
        )
    )
    progsnap_events: str = field(
        default="run",
        metadata=metadata(
            help="Choose what level of event to capture from Progsnap event logs.",
            choices=["run", "edit", "last", "compile"]
        )
    )
    cache: Optional[str] = field(
        default=None,
        metadata=metadata(
            help="Use the given directory to hold the cache."
                 " You can use './' to use the current directory.",
        )
    )
    threaded: bool = field(
        default=False,
        metadata=metadata(
            help="Run the instructor script in a separate thread to avoid"
                 " timeout crashes.",
            action="store_true"
        )
    )
    log_level: str = field(
        default="ERROR",
        metadata=metadata(
            help="Set the logging level for Pedal.",
            choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        )
    )

    tool: Dict[str, Dict[str, Any]] = field(
        default_factory=dict,
        metadata=metadata(
            help="Additional configuration for specific tools. Provide"
                        " them in the format `--tool name.setting=value`, e.g.,"
                        " `--tool tifa.skip=true`.",
            nargs="*"
        )
    )

    points: str = field(
        default="1",
        metadata=metadata(
            help="The maximum number of points to give for a correct answer. Use the value 'calculate' to automatically calculate the points from the grading script.",
        )
    )

    pool: str = field(
        default=False,
        metadata=metadata(
            help="Whether or not to use the pool system, which allows for"
                 " randomization of questions and instructions in the"
                 " instructor control script. If set to 'comment', then"
                 " the ICS will be divided into pools based on comments, and the "
                 " pool will be chosen based on the presence of a comment in the student code."
        )
    )

    execution: Dict[str, Any] = field(
        default_factory=dict,
        metadata=metadata(
            help="Additional metadata that will be passed to the submission as `execution`. Provide"
                        " them in the format `--execution setting=value`, e.g.,"
                        " `--execution submission_id=27`.",
        )
    )

def parse_nested_dict(items, minimum_key_parts=2):
    """Convert ['sandbox.trace=true', 'tifa.allow_none_objects=true'] into a nested dict"""
    nested_dict = defaultdict(dict)
    print(items)
    for item in items:
        if '=' not in item:
            raise ValueError(f"Invalid format: {item}. Expected key=value.")
        key, value = item.split('=', 1)
        parts = key.split('.')
        if len(parts) < minimum_key_parts:
            raise ValueError(f"Invalid tool key: {key}. Must be in format 'name.setting'.")
        if len(parts) >= 2:
            tool_name, setting = parts[0], '.'.join(parts[1:])
            nested_dict[tool_name][setting] = value
        else:
            setting = parts[0]
            nested_dict[setting] = value
    return dict(nested_dict)


def parse_config_file(path):
    config_data = parse_out_file(path)
    defaults = config_data.pop("default")

    for section_name, items in config_data.items():
        if section_name.startswith("tool."):
            tool_name = section_name[5:]
            if not tool_name:
                raise ValueError("Tool name cannot be empty.")
            defaults[tool_name] = items
        else:
            raise ValueError(f"Unknown config section: {section_name}")

    return defaults



def make_job_config_parser():
    # Root parser, that will be extended with the real parser.
    # This allows us to extract out the config file first.
    config_parser = argparse.ArgumentParser(description=MAIN_PARSER_DESCRIPTION,
                                            add_help=False)
    config_parser.add_argument("--config", "-c",
                               help="Uses the configuration file to get settings.")
    args, remaining_argv = config_parser.parse_known_args()

    defaults = {}
    if args.config:
        # Parse the config file, update the defaults accordingly
        defaults.update(parse_config_file(args.config))

    parser = argparse.ArgumentParser(description=MAIN_PARSER_DESCRIPTION,
                                     parents=[config_parser])

    fields = [field for field in JobConfig.__dataclass_fields__.items()
              if field[0] != "config"]
    for field_name, field_obj in fields:
        meta = field_obj.metadata

        names = ["--" + field_name] if meta["optional"] else [field_name]
        if meta["short_form"]:
            names.append("-" + meta["short_form"])
        default_function = field_obj.default_factory() if callable(field_obj.default_factory) else field_obj.default
        parameters = {
            "help": meta["help"],
            "action": meta["action"]
        }
        if "choices" in meta:
            parameters["choices"] = meta["choices"]
        if "nargs" in meta:
            parameters["nargs"] = meta["nargs"]
        parser.add_argument(*names,
                            default=default_function,
                            **parameters
                            )

    parser.set_defaults(**defaults)

    # Reset `required` attribute when provided from config file
    for action in parser._actions:
        if action.dest in defaults:
            action.required = False

    parsed_args = parser.parse_args(remaining_argv)

    # Post process
    parsed_args.tool = parse_nested_dict(parsed_args.tool or [])
    parsed_args.execution = parse_nested_dict(parsed_args.execution or [], minimum_key_parts=1)

    return JobConfig(**vars(parsed_args))




from pedal.command_line import run_job
from pedal.command_line.modes import StatsPipeline, MODES
from pedal.core.config_job import JobConfig


def test_stats_pipeline():
    """
    Test the StatsPipeline class to ensure it initializes correctly and can run.
    """
    pipeline = StatsPipeline(JobConfig(
        mode=MODES.STATS,
        submissions = "def main():\n    return 'Hello, World!'",
        instructor = "from pedal import *\nensure_function('main', score='+60%')",
        instructor_direct = True,
        submission_direct = True,
        points= ".5",
        output=None
    ))
    assert pipeline is not None, "StatsPipeline should be initialized"

    # Run the pipeline and check if it completes without errors
    try:
        result = pipeline.execute()
    except Exception as e:
        assert False, f"StatsPipeline run failed with exception: {e}"

    # Check if the report is generated
    assert pipeline.submissions[0].result.resolution.score == .5, "Expected score to be 1.5"

    print(pipeline.result)


def test_run_job():
    """
    Test the run_job function with StatsPipeline mode.
    """
    on_run = "from pedal import *\nsuppress('algorithmic')\nensure_function('main', score='+60%')\n#### pool_1 ####\nensure_ast('For')\n#### pool_2 ####\nensure_ast('If')\n#### Footer ####\n"
    student_code = "def main():\n    return 'Hello, World!'"
    submissions = run_job(
        submissions = student_code,
        instructor = on_run,
        instructor_direct = True,
        submission_direct = True,
        points= ".5",
        pool = "submission_id",
        execution = {"submission_id": 0},
    )

    assert submissions is not None, "run_job should return a result"
    assert submissions[0].result.error is None
    assert submissions[0].result.resolution.score == .5, "Expected score to be 1.5"
    assert submissions[0].result.resolution.label == 'ensure_ast'
    assert submissions[0].result.resolution.category == 'specification'
    assert submissions[0].result.resolution.data['name'] == 'For'
    if not "MAIN_REPORT" in submissions[0].result.data:
        raise submissions[0].result.error
    main_report = submissions[0].result.data['MAIN_REPORT']
    print(main_report)

    submissions = run_job(
        submissions=student_code,
        instructor=on_run,
        instructor_direct=True,
        submission_direct=True,
        points=".5",
        pool="submission_id",
        execution={"submission_id": 47},
    )

    assert submissions is not None, "run_job should return a result"
    assert submissions[0].result.error is None
    assert submissions[0].result.resolution.score == .5, "Expected score to be 1.5"
    assert submissions[0].result.resolution.label == 'ensure_ast'
    assert submissions[0].result.resolution.data['name'] == 'If'
    assert submissions[0].result.resolution.category == 'specification'
    if not "MAIN_REPORT" in submissions[0].result.data:
        raise submissions[0].result.error
    main_report = submissions[0].result.data['MAIN_REPORT']


"""
Simple enumeration of the Modes available for Pedal's command line.
"""
import re
import json
import os
import sys
import traceback
from contextlib import redirect_stdout
from io import StringIO
import unittest
from textwrap import indent
from unittest.mock import patch
from pprint import pprint
import warnings
import argparse
from collections import Counter

from pedal.command_line.report import StatReport
from pedal.command_line.verify import generate_report_out, ReportVerifier
from pedal.core.report import MAIN_REPORT
from pedal.core.submission import Submission
from pedal.utilities.files import normalize_path, find_possible_filenames
from pedal.utilities.progsnap import SqlProgSnap2
from pedal.utilities.text import chomp
from pedal.resolvers.export import PedalJSONEncoder, clean_json

try:
    from tqdm import tqdm
except ImportError:
    def tqdm(values):
        """ Trivial helper function to replace TQDM's behavior """
        yield from values


# Disable all warnings from student files
# TODO: Move this strictly into ICS and Submission execution
warnings.filterwarnings("ignore")


def is_progsnap(path):
    """ Determines if this is a ProgSnap file or folder. """
    if os.path.isfile(path):
        name, extension = os.path.splitext(path)
        return extension in ('.zip', '.db')
    else:
        # Does it have all the necessary components?
        return os.path.isfile(os.path.join(path, 'MainTable.csv'))


def get_python_files(paths):
    result = {}
    for path in paths:
        name, ext = os.path.splitext(path)
        if ext != ".py":
            continue
        with open(path, 'r') as submission_file:
            contents = submission_file.read()
        result[path] = contents
    return result


class BundleResult:
    """
    Represents the result of running an instructor control script on a submission.
    This includes not only the output and error, but also the resolution of the feedback (aka
    the final feedback). Also includes the data that was generated during the execution.
    """
    def __init__(self, data, output, error, resolution):
        self.data = data
        self.output = output
        self.error = error
        self.resolution = resolution

    def to_json(self):
        resolution = self.resolution.copy() if self.resolution else {}
        if not isinstance(resolution, dict):
            resolution = resolution.to_json()
        #if 'considered' in resolution:
        #    for c in resolution['considered']:
        #        if 'fields' in c:
        #            del c['fields']
        return dict(
            output=self.output,
            error=self.error,
            **resolution
        )

def extract_instructor_pool(config, script, submission, pool_mode='comment'):
    parts = re.split(r"#{4,} (.+) #{4,}", script, flags=re.M)
    unnamed_parts = re.split(r"#{4,} .+ #{4,}", script, flags=re.M)
    if len(parts) <= 2:
        raise ValueError(
            "Pools not found in instructor control script. Please specify a pool in the instructor control script by writing `#### <pool_name> ####` at the top of the file. Note the hash signs and spaces are necessary, but you do not need the angle brackets."
        )
    header, footer = parts[0], parts[-1]

    if pool_mode == 'comment':
        for line in submission.main_code.splitlines():
            if line.startswith("# pool: "):
                pool_name = line[8:].strip()
                break
        else:
            raise ValueError(
                "No pool found in submission. Please specify a pool in the submission file by writing `# pool: <pool_name>` at the top of the file."
            )
        if not pool_name:
            raise ValueError(
                "No pool found in submission. Please specify a pool in the submission file by writing `# pool: <pool_name>` at the top of the file."
            )
        for i in range(1, len(parts) - 1, 2):
            if parts[i] == pool_name:
                body = parts[i + 1]
                break
        else:
            found_pools = ", ".join(parts[1::2])
            raise ValueError(
                f"Pool '{pool_name}' not found in instructor control script. Please specify a pool in the instructor control script by writing `#### {pool_name} ####` at the top of the file. Note the hash signs and spaces are necessary, but you do not need the angle brackets. Found pools: {found_pools}."
            )
    elif pool_mode == 'submission_id':
        pool_id = submission.execution.get('submission_id') if submission.execution else None
        if pool_id is None:
            raise ValueError(
                "No pool id found in submission. Please ensure the submission has a submission_id for the execution."
            )
        pool_index = 1 + (pool_id % (len(unnamed_parts)-2))
        body = unnamed_parts[pool_index]
    else:
        raise ValueError(f"Unknown pool mode: {pool_mode}")
    # print(header, body, footer)
    return "\n".join([ header, body, footer ])

class Bundle:
    """
    Represents the combination of an instructor control script and a submission that it is
    being run on. Also includes the environment that the script is being run in, and the
    result of the execution (once the execution is finished). Finally, also includes the
    configuration that was used to run the bundle.
    """
    def __init__(self, config, script, submission):
        if config.pool in ('comment', 'submission_id'):
            script = extract_instructor_pool(config, script, submission, config.pool)
        self.config = config
        self.script = script
        self.submission = submission
        self.environment = None
        self.result = None

    def to_json(self):
        return dict(
            script=self.script,
            submission=self.submission.to_json(),
            environment=self.environment,
            result=self.result.to_json()
        )

    def run_ics_bundle(self, resolver='resolve', skip_tifa=False, skip_run=False):
        """
        Runs the given instructor control script on the given submission, with the
        accompany contextualizations.
        """
        ics_args = [self.submission, self.environment]
        captured_output = StringIO()
        global_data = {}
        error = None
        resolution = None
        if self.environment:
            env = __import__('pedal.environments.' + self.environment,
                             fromlist=[''])
            global_data.update(env.setup_environment(self.submission,
                                                     skip_tifa=skip_tifa,
                                                     skip_run=skip_run,
                                                     threaded=self.config.threaded).fields)
        else:
            # TODO: Check that this shouldn't be for global_data instead of MAIN_REPORT directly
            MAIN_REPORT.contextualize(self.submission)
        if self.config.points:
            MAIN_REPORT.set_max_points(self.config.points)
        with redirect_stdout(captured_output):
            with patch.object(sys, 'argv', ics_args):
                try:
                    grader_exec = compile(self.script,
                                          self.submission.instructor_file, 'exec')
                    exec(grader_exec, global_data)
                    if 'MAIN_REPORT' in global_data:
                        if not global_data['MAIN_REPORT'].resolves:
                            if resolver in global_data:
                                resolution = global_data[resolver]()
                            # TODO: Need more elegance/configurability here
                            elif self.config.resolver == 'resolve':
                                exec("from pedal.resolvers import print_resolve", global_data)
                                resolution = global_data["print_resolve"]()
                        else:
                            resolution = global_data['MAIN_REPORT'].resolves[-1]
                except Exception as e:
                    error = e
        actual_output = captured_output.getvalue()
        self.result = BundleResult(global_data, actual_output, error, resolution)


class AbstractPipeline:
    """
    Generic pipeline for handling all the phases of executing instructor
    control scripts on submissions, and reformating the output.
    Should be subclassed instead of used directly.
    """

    def __init__(self, config):
        if isinstance(config, dict):
            # TODO: Include default argument automatically
            config = argparse.Namespace(**config)
        self.config = config
        self.submissions = []
        self.result = None

    def execute(self):
        self.load_submissions()
        self.setup_execution()
        self.run_control_scripts()
        return self.process_output()

    def load_file_submissions(self, scripts):
        # Get instructor control scripts
        all_scripts = []
        for script in scripts:
            script_file_name, script_file_extension = os.path.splitext(script)
            # Single Python file
            if script_file_extension in ('.py',):
                with open(script, 'r') as scripts_file:
                    scripts_contents = scripts_file.read()
                all_scripts.append((script, scripts_contents))
        given_submissions = self.config.submissions
        # If submission is a directory, use it as a directory adjacent to each ics
        if os.path.isdir(given_submissions):
            for script, scripts_contents in all_scripts:
                directory_pattern = given_submissions
                submission_dir = normalize_path(directory_pattern, script)
                submission_files = [
                    os.path.join(submission_dir, sub)
                    for sub in os.listdir(submission_dir)
                ]
                subs = get_python_files(submission_files)
                for main_file, main_code in subs.items():
                    new_submission = Submission(
                        main_file=main_file, main_code=main_code,
                        instructor_file=script,
                        execution=self.config.execution
                    )
                    self.submissions.append(Bundle(self.config, scripts_contents, new_submission))
        # Otherwise, if the submission is a single file:
        # Maybe it's a Progsnap DB file?
        elif given_submissions.endswith('.db'):
            for script, scripts_contents in all_scripts:
                self.load_progsnap(given_submissions, instructor_code=scripts_contents)
        # Otherwise, must just be a single python file.
        else:
            main_file = given_submissions
            load_error, possible_load_error = None, None
            alternatives = [given_submissions]
            # if alternative filenames given, we'll queue them up
            if self.config.alternate_filenames:
                alternatives.extend(find_possible_filenames(self.config.alternate_filenames))
            # Run through all possible filenames
            for possible in alternatives:
                try:
                    with open(possible, 'r') as single_submission_file:
                        main_code = single_submission_file.read()
                        main_file = possible
                    break
                except OSError as e:
                    # Only capture the first possible load error
                    if possible_load_error is None:
                        possible_load_error = e
            else:
                # Okay, file does not exist. Load error gets triggered.
                main_code = None
                load_error = possible_load_error
            for script, scripts_contents in all_scripts:
                new_submission = Submission(
                    main_file=main_file, main_code=main_code,
                    instructor_file=script, load_error=load_error,
                    execution=self.config.execution
                )
                self.submissions.append(Bundle(self.config, scripts_contents, new_submission))
            return load_error

    progsnap_events_map = {
        'run': 'Run.Program',
        'compile': 'Compile',
        'edit': 'File.Edit',
        'last': 'File.Edit'
    }

    def load_progsnap(self, path, instructor_code=None):
        script_file_name, script_file_extension = os.path.splitext(path)
        if script_file_extension in ('.db',):
            with SqlProgSnap2(path, cache=self.config.cache) as progsnap:
                if self.config.progsnap_profile:
                    progsnap.set_profile(self.config.progsnap_profile)
                link_filters = {}
                include_scripts = self.config.include_scripts
                if include_scripts:
                    link_filters['Assignment'] = {}
                    if include_scripts.startswith('name='):
                        link_filters['Assignment']['X-Name'] = include_scripts[5:]
                    elif include_scripts.startswith('id='):
                        link_filters['Assignment']['AssignmentId'] = include_scripts[3:]
                    elif include_scripts.startswith('url='):
                        link_filters['Assignment']['X-URL'] = include_scripts[4:]
                    else:
                        link_filters['Assignment']['X-URL'] = include_scripts
                event_type = self.progsnap_events_map[self.config.progsnap_events]
                events = progsnap.get_events(event_filter={'EventType': event_type},
                                             link_filters=link_filters, limit=self.config.limit)
            if self.config.progsnap_events == 'last':
                runs_by_user_assign = {}
                for event in sorted(events, key=lambda e: e['event_id']):
                    key = (event[progsnap.PROFILES[progsnap.profile]['link_primary']['user']], event['assignment_name'])
                    runs_by_user_assign[key] = event
                events = list(runs_by_user_assign.values())
            for event in events:
                if instructor_code is None:
                    instructor_code_for_this_run = event['on_run']
                else:
                    instructor_code_for_this_run = instructor_code
                link_selections = progsnap._merge('link_selections', {})
                new_submission = Submission(
                    main_file='answer.py',
                    main_code=event['submission_code'] if isinstance(event['submission_code'], str) else
                        event['submission_code'].decode('utf-8'),
                    instructor_file='instructor.py',
                    #files={'cisc106.py': 'from bakery import *'},
                    execution=dict(client_timestamp=event['client_timestamp'],
                                   event_id=event['event_id']),
                    #user=dict(email=event['student_email'],
                    #          first=event['student_first'],
                    #          last=event['student_last']),
                    user={key: event[key] for key in link_selections['Subject'].values()}
                    if 'Subject' in link_selections else {'id': event['subject_id']},
                    assignment=dict(name=event['assignment_name'],
                                    url=event['assignment_url']),
                )
                self.submissions.append(Bundle(self.config, instructor_code_for_this_run, new_submission))
            # raise ValueError("TODO: ProgSnap DB files not yet supported")
            # Progsnap Zip
        elif script_file_extension in ('.zip',):
            raise ValueError("TODO: Zip files not yet supported")

    def load_file_directly(self, config):
        new_submission = Submission(
            main_file="answer.py",
            main_code=self.config.submissions,
            instructor_file=self.config.instructor,
            execution=config.execution
        )
        self.submissions.append(Bundle(self.config, self.config.instructor, new_submission))

    def load_submissions(self):
        given_script = self.config.instructor
        if self.config.instructor_direct:
            # TODO: Allow non-progsnap ics_direct
            if self.config.submission_direct:
                self.load_file_directly(self.config)
            else:
                self.load_progsnap(self.config.submissions, instructor_code=given_script)
        elif is_progsnap(given_script):
            self.load_progsnap(given_script)
        elif os.path.isfile(given_script):
            self.load_file_submissions([given_script])
        elif os.path.isdir(given_script):
            python_files = os.listdir(given_script)
            self.load_file_submissions(python_files)
        else:
            potential_filenames = list(find_possible_filenames(given_script))
            for filename in potential_filenames:
                load_error = self.load_file_submissions([filename])
                if load_error is None:
                    return
            from pedal.source.feedbacks import source_file_not_found
            source_file_not_found(potential_filenames[0], False)

    def setup_execution(self):
        for bundle in self.submissions:
            bundle.environment = self.config.environment

    def run_control_scripts(self):
        for bundle in self.submissions:
            bundle.run_ics_bundle(resolver=self.config.resolver,
                                  skip_tifa=self.config.skip_tifa,
                                  skip_run=self.config.skip_run)

    def process_output(self):
        for bundle in self.submissions:
            print(bundle.submission.instructor_file,
                  bundle.submission.main_file,
                  bool(bundle.result.error))


class FeedbackPipeline(AbstractPipeline):
    """
    ``feedback``: Pipeline for running the instructor control script on a submission and
    then printing the resolver output to the console. Often the most useful
    if you are trying to deliver the feedback without a grade.
    """
    def process_output(self):
        for bundle in self.submissions:
            #print(bundle.submission.instructor_file,
            #      bundle.submission.main_file)
            if bundle.result.error:
                print(bundle.result.error)
            elif bundle.result.resolution:
                # report = bundle.result.data['MAIN_REPORT']
                #print(bundle.result.resolution.title)
                #print(bundle.result.resolution.message)
                print(bundle.result.output)
            else:
                print("Error: No feedback determined.")


class RunPipeline(AbstractPipeline):
    """
    ``run``: Pipeline for running the instructor control script on a submission and generating a report
    file in the `ini` file format. This is a simple file format that has a lot of the interesting
    fields. The file is not actually dumped to the filesystem, but instead printed directly.
    So this is a good way to run students' code in a sandbox and see what comes out.
    """
    def process_output(self):
        for bundle in self.submissions:
            #print(bundle.submission.instructor_file,
            #      bundle.submission.main_file)
            if bundle.result.error:
                print(bundle.result.error)
            elif bundle.result.resolution:
                report = bundle.result.data['MAIN_REPORT']
                generate_report_out(None, self.config.environment, report)
            else:
                print("No feedback determined.")


class StatsPipeline(AbstractPipeline):
    """
    ``stats``: Pipeline for running the instructor control script on a submission and then
    dumping a JSON report with all the feedback objects. This is useful for
    analyzing the feedback objects in a more programmatic way.
    """
    def run_control_scripts(self):
        for bundle in tqdm(self.submissions):
            bundle.run_ics_bundle(resolver='stats_resolve', skip_tifa=self.config.skip_tifa,
                                  skip_run=self.config.skip_run)

    def process_output(self):
        total = 0
        errors = 0
        final = []
        for bundle in self.submissions:
            if bundle.result.error:
                print(bundle.result.error)
                errors += 1
            else:
                final.append(bundle.to_json())
            total += 1
        final = clean_json(final)
        if self.config.output is not None:
            #print(final)
            print("Total Processed:", total)
            print("Errors:", errors)

            feedback_by_label_category = Counter()
            scored_feedback_by_label_category = Counter()
            for bundle in self.submissions:
                final = bundle.result.resolution
                feedback_by_label_category[(final.category, final.label)] += 1
                for feedback in final._scores_feedback:
                    scored_feedback_by_label_category[(feedback.category, feedback.label)] += 1
            print("Final Feedback by label/category:")
            for (category, label), count in feedback_by_label_category.items():
                print(f"  {category} - {label}: {count}")

            print("Scored Feedback by label/category:")
            for (category, label), count in scored_feedback_by_label_category.items():
                print(f"  {category} - {label}: {count}")


            # pedal_json_encoder = PedalJSONEncoder(indent=2, skipkeys=True)
            # if self.config.output == 'stdout':
            #     print(pedal_json_encoder.encode(final))
            # else:
            #     with open(self.config.output, 'w') as output_file:
            #         print(pedal_json_encoder.encode(final), file=output_file)
        return StatReport(final)



class VerifyPipeline(AbstractPipeline):
    """
    ``verify``: Pipeline for running the instructor control script on a submission and then
    comparing the output to an expected output file. This is useful for verifying
    that the feedback is correct (at least, as correct as the expected output).

    You can also use this pipeline to generate the output files, to quickly create
    regression "tests" of your feedback scripts.
    """
    def process_output(self):
        for bundle in self.submissions:
            bundle.run_ics_bundle(resolver=self.config.resolver, skip_tifa=self.config.skip_tifa,
                                  skip_run=self.config.skip_run)
            if bundle.result.error:
                raise bundle.result.error
            # Now get the expected output
            base = os.path.splitext(bundle.submission.main_file)[0]
            test_name = os.path.basename(base)
            if os.path.exists(base + ".out"):
                verifier = ReportVerifier(base + ".out", self.config.environment)
                self.add_case(test_name, bundle.result, verifier)
            else:
                if self.config.create_output:
                    report = bundle.result.data['MAIN_REPORT']
                    generate_report_out(base + ".out", self.config.environment, report)
                    self.skip_case(test_name)
                else:
                    self.error_case(test_name, bundle.result)
        self.run_cases()

    def __init__(self, config):
        super().__init__(config)

        class TestReferenceSolutions(unittest.TestCase):
            __module__ = 'VerifyTestCase'
            __qualname__ = config.instructor_name
            maxDiff = None

            def __str__(self):
                return f"{config.instructor}, using {self._testMethodName[5:]}.py"
                # return f"{self.id().__name__} (Test{config.instructor})"

            def __repr__(self):
                return str(self)

        self.TestReferenceSolutions = TestReferenceSolutions
        self.tests = []

    def add_case(self, name, bundle_result, verifier):
        """ Add this ``name`` as a new test. """
        if bundle_result.output.strip():
            print(bundle_result.output)
        error = bundle_result.error if bundle_result.error else None
        if error:
            print(error)
        if 'MAIN_REPORT' not in bundle_result.data:
            # TODO: Better error message here
            raise Exception("No MAIN_REPORT found; did you import Pedal?")
        actual = bundle_result.data['MAIN_REPORT'].result.to_json()
        error = bundle_result.error
        expected_fields = verifier.get_final()

        def new_test(self):
            if error:
                raise error
            entire_actual, entire_expected = "", ""
            differing_fields = []
            for field, value in expected_fields:
                self.assertIn(field, actual, msg="Field was not in feedback.")
                actual_value = chomp(str(actual[field]).strip())
                expected_value = chomp(str(value))
                entire_actual += field + ": " + actual_value + "\n"
                entire_expected += field + ": " + expected_value + "\n"
                if expected_value != actual_value:
                    differing_fields.append(field)
            differing_fields = ', '.join(f"'{field}'" for field in differing_fields)
            self.assertEqual(entire_expected, entire_actual,
                             msg=f"Wrong value for {differing_fields} in {name}.")

        setattr(self.TestReferenceSolutions, "test_" + name, new_test)
        self.tests.append("test_" + name)

    def skip_case(self, name):
        def new_test(self):
            self.skipTest("Output did not exist; created.")

        setattr(self.TestReferenceSolutions, "test_" + name, new_test)
        self.tests.append("test_" + name)

    def error_case(self, name, bundle_result):
        def new_test(self):
            self.fail(
                "Expected output file was not found next to the instructor file. Perhaps you meant to use --create_output?")

        setattr(self.TestReferenceSolutions, "test_" + name, new_test)
        self.tests.append("test_" + name)

    def run_cases(self):
        suite = unittest.defaultTestLoader.loadTestsFromTestCase(self.TestReferenceSolutions)
        runner = unittest.TextTestRunner()
        runner.run(suite)


class GradePipeline(AbstractPipeline):
    """
    ``grade``: Pipeline for running the instructor control script on a submission and then outputing
    the grade to the console. This is useful for quickly grading a set of submissions.
    The instructor file, student data, and assignment are also all printed out in the following CSV format:

    instructor_file, student_file, student_email, assignment_name, score, correct
    """
    def process_output(self):
        if self.config.output == 'stdout':
            self.print_bundles(sys.stdout)
        else:
            with open(self.config.output, 'w') as output_file:
                self.print_bundles(output_file)

    def print_bundles(self, target):
        #print(len(self.submissions))
        for bundle in self.submissions:
            # print(bundle.result.output, file=target)
            if bundle.result.error:
                raise bundle.result.error
            # This info is not sent to the output target, just to stdout
            print(bundle.submission.instructor_file,
                  bundle.submission.main_file,
                  bundle.submission.user.get('student_email') if bundle.submission.user else 'Unknown User',
                  bundle.submission.assignment.get('name') if bundle.submission.assignment else 'Unknown Assignment',
                  1 if bundle.result.resolution.correct else
                  bundle.result.resolution.score,
                  sep=", ", end="\n")


class SandboxPipeline(AbstractPipeline):
    """
    ``sandbox``: Pipeline for running ONLY the student's code, and then outputing the results to the console.
    There is no instructor control script logic, although the Source tool does check that the
    student's code is syntactically correct. Otherwise, the students' code is run in a Sandbox mode.
    This is useful if you just want to safely execute student code and observe their output.
    """

    ICS = """from pedal import *
verify()
run()"""

    def load_submissions(self):
        # Use first argument as student submissions
        given_script = self.config.instructor
        # ... either a single file
        if os.path.isfile(given_script):
            scripts = [given_script]
        # ... or a directory of files
        else:
            scripts = os.listdir(given_script)
        # Then create submission to run
        for script in scripts:
            script_file_name, script_file_extension = os.path.splitext(script)
            if script_file_extension in ('.py',):
                load_error = None
                try:
                    with open(script, 'r') as scripts_file:
                        scripts_contents = scripts_file.read()
                except Exception as e:
                    load_error = e
                new_submission = Submission(
                    main_file="answer.py", main_code=scripts_contents,
                    instructor_file=script, load_error=load_error
                )
                self.submissions.append(Bundle(self.config, self.ICS, new_submission))

    def process_output(self):
        # Print output
        # Print runtime exceptions, if any
        for bundle in self.submissions:
            if bundle.result.error:
                traceback.print_tb(bundle.result.error.__traceback__)
                print(bundle.result.error)
                continue
            sandbox = bundle.result.data['MAIN_REPORT']['sandbox']['sandbox']
            output = [sandbox.raw_output]
            if sandbox.feedback:
                output.append(sandbox.feedback.message)
            print("\n".join(output))


class DebugPipeline(AbstractPipeline):
    """
    ``debug``: Pipeline for running the instructor control script on a submission and then outputing
    the full results to the console. This is useful for debugging the instructor control
    script, as it will show the full output, error, all of the feedback objects considered,
    and the final feedback.
    """
    def process_output(self):
        for bundle in self.submissions:
            print(bundle.submission.instructor_file,
                  bundle.submission.main_file)
            print("****** Student Code:")
            print(indent(bundle.submission.main_code, "    "))
            print("****** Results")
            if bundle.result.error:
                print(bundle.result.error)
            else:
                print("Output:")
                print(indent(bundle.result.output, " " * 4))
                report = bundle.result.data['MAIN_REPORT']
                print("Feedback:")
                for feedback in report.feedback:
                    print("\t", feedback, repr(feedback.fields))
                print("Final Feedback")
                pprint(report.result.to_json())


class MODES:
    """
    The possible modes that Pedal can be run in.
    """
    # Runs the instructor control script and outputs all feedback
    RUN = 'run'
    # Get just the feedback for submissions
    FEEDBACK = 'feedback'
    # Get just the score/correctness
    GRADE = 'grade'
    # Check against expected output
    VERIFY = 'verify'
    # Output all the feedback objects
    STATS = 'stats'
    # Run the file as student code using the sandbox
    SANDBOX = 'sandbox'
    # Output as much information as possible
    DEBUG = 'debug'

    PIPELINES = {
        RUN: RunPipeline,
        FEEDBACK: FeedbackPipeline,
        GRADE: GradePipeline,
        STATS: StatsPipeline,
        SANDBOX: SandboxPipeline,
        VERIFY: VerifyPipeline,
        DEBUG: DebugPipeline
    }




import argparse
from typing import Optional, List, Dict, Any, Type, TypeVar
from pedal.core.report import Report
from pedal.core.resolver import Resolver
from pedal.environments import ALL_ENVIRONMENTS
from pedal.core.config_job import JobConfig
from dataclasses import dataclass, field

T = TypeVar("T", bound="EnvironmentConfig")

@dataclass
class EnvironmentConfig:
    job_config: JobConfig
    files: Dict[str, str]
    main_file: str
    main_code: str
    instructor_file: str
    # Metadata
    user: Any
    assignment: Any
    course: Any
    execution: Any
    # Legacy Tool settings
    skip_tifa: bool
    skip_run: bool
    real_io: bool
    path_mask: Optional[str]
    inputs: List[str]
    trace: bool
    threaded: bool
    # New Style Tool Settings
    tool: Dict[str, Dict]  # TODO: Make this happen
    # Resolver controls
    set_correct: bool
    set_success: bool
    resolver: Optional[Resolver] # TODO: Make this happen
    report: Report


import pedal.types.library.matplotlib
import pedal.types.library.cs1_curriculum
import pedal.types.library.designer
import pedal.types.library.drafter
import pedal.types.library.turtles
import pedal.types.library.microbit
import pedal.types.library.pillow



from pedal.core.commands import compliment, explain, gently, give_partial, suppress
from pedal.core.feedback_category import FeedbackCategory
from pedal.core.report import MAIN_REPORT
from pedal.sandbox.commands import run
from pedal.assertions.runtime import *
from pedal.assertions.static import ensure_function


class QuestionGrader:
    def _get_functions_with_filter(self, filter='grade_'):
        return [getattr(self, method_name) for method_name in dir(self)
                if method_name.startswith(filter) and
                callable(getattr(self, method_name))]

    def _test(self, question):
        methods = self._get_functions_with_filter()
        return [method(question) for method in methods]


class FunctionGrader(QuestionGrader):
    """

    """
    MAX_POINTS = 10
    DEFINITION_POINTS = 3
    COMPONENTS_POINTS = 1
    MAX_COMPONENTS_POINTS = 2
    UNIT_TEST_TYPE_POINTS = None
    UNIT_TEST_VALUE_POINTS = None
    UNIT_TEST_TOTAL_POINTS = 5
    UNIT_TEST_TYPE_RATIO = .5
    UNIT_TEST_COMPLETION_POINTS = 2

    def __init__(self, function_name, signature, tests, config=None,
                 max_points=None, definition_points=None, components_points=None,
                    max_components_points=None, unit_test_type_points=None,
                    unit_test_value_points=None, unit_test_total_points=None,
                    unit_test_type_ratio=None, unit_test_completion_points=None):
        super().__init__()
        self.function_name = function_name
        self.signature = signature
        self.tests = tests
        self.points = 0
        self.config = config or {}
        self.config.setdefault('suppress_function_unused', True)
        self.config.setdefault('exact_strings', False)
        self.justification_parts = []
        if max_points is not None:
            self.MAX_POINTS = max_points
        if definition_points is not None:
            self.DEFINITION_POINTS = definition_points
        if components_points is not None:
            self.COMPONENTS_POINTS = components_points
        if max_components_points is not None:
            self.MAX_COMPONENTS_POINTS = max_components_points
        if unit_test_type_points is not None:
            self.UNIT_TEST_TYPE_POINTS = unit_test_type_points
        if unit_test_value_points is not None:
            self.UNIT_TEST_VALUE_POINTS = unit_test_value_points
        if unit_test_total_points is not None:
            self.UNIT_TEST_TOTAL_POINTS = unit_test_total_points
        if unit_test_type_ratio is not None:
            self.UNIT_TEST_TYPE_RATIO = unit_test_type_ratio
        if unit_test_completion_points is not None:
            self.UNIT_TEST_COMPLETION_POINTS = unit_test_completion_points

    def _test(self, question):
        defined = self.grade_definition(question)

        if not defined:
            return self.report_status(question)

        self.grade_components(question)

        passed_tests = self.grade_unit_tests(question)
        if not passed_tests:
            return self.report_status(question)

        self.report_success(question)

    def report_status(self, question):
        """

        Args:
            question:
        """
        self.justification_parts.append(f"Total: {self.points:.2f}/{self.MAX_POINTS} (+{round(self.points/self.MAX_POINTS * 100)}%)")
        give_partial(self.points/self.MAX_POINTS,
                     title="Function Grader Partial Credit",
                     message="Partial credit from function grader",
                     justification="\n".join(self.justification_parts))

    def report_success(self, question):
        """

        Args:
            question:
        """
        question.answer()

    def grade_definition(self, question):
        """

        Args:
            question:

        Returns:

        """
        self.student = run()

        if self.config.get('suppress_function_unused', True):
            suppress(FeedbackCategory.ALGORITHMIC, 'unused_variable', {'name': self.function_name})

        if ensure_function(self.function_name, *self.signature):
            if not ensure_function(self.function_name, muted=True):
                gently(f"The function {self.function_name} has an incorrect signature; check the function name, parameter names and types, and the return type.")
            else:
                gently(f"The function {self.function_name} was not defined.")
            return False

        if self.student.exception:
            return False

        if assertHasFunction(self.student, self.function_name):
            gently("Function defined incorrectly")
            return False



        self.points += self.DEFINITION_POINTS
        self.justification_parts.append(f"Function {self.function_name} defined with signature: {self.DEFINITION_POINTS}")
        return True

    def grade_components(self, question):
        """

        Args:
            question:
        """
        self.component_points = 0
        components = self._get_functions_with_filter('grade_component_')
        for component in components:
            component(question)
        self.component_points = min(self.component_points, self.MAX_COMPONENTS_POINTS)
        self.points += self.component_points
        if self.component_points:
            self.justification_parts.append(f"Components: {self.component_points}")

    def assertEqual(self, *parameters):
        """

        Args:
            *parameters:

        Returns:

        """
        return assertEqual(*parameters, exact_strings=self.config.get('exact_strings', False))

    def grade_unit_tests(self, question):
        """

        Args:
            question:

        Returns:

        """
        all_good = True
        if self.UNIT_TEST_TOTAL_POINTS is None:
            TYPE_POINT_ADD = self.UNIT_TEST_TYPE_POINTS
            VALUE_POINT_ADD = self.UNIT_TEST_VALUE_POINTS
        else:
            ratio = self.UNIT_TEST_TYPE_RATIO
            TYPE_POINT_ADD = (self.UNIT_TEST_TOTAL_POINTS / len(self.tests) * ratio)
            VALUE_POINT_ADD = (self.UNIT_TEST_TOTAL_POINTS / len(self.tests) * (1 - ratio))
        test_points = 0
        for index, (arguments, expected) in enumerate(self.tests):
            # import sys
            # print(repr(arguments), file=sys.stderr)
            result = self.student.call(self.function_name, *arguments)
            # print(repr(self.student.exception), file=sys.stderr)
            if self.student.exception:
                all_good = False
                self.justification_parts.append(f"  Test {index}: error occurred")
                continue
            if not assertIsInstance(result, type(expected)):
                self.points += TYPE_POINT_ADD
                test_points += TYPE_POINT_ADD
            else:
                all_good = False
                self.justification_parts.append(f"  Test {index}: wrong return type")
                continue
            if not self.assertEqual(result, expected):
                self.points += VALUE_POINT_ADD
                test_points += VALUE_POINT_ADD
                self.justification_parts.append(f"  Test {index}: correct ({VALUE_POINT_ADD})")
            else:
                all_good = False
                self.justification_parts.append(f"  Test {index}: wrong return value, but correct type ({TYPE_POINT_ADD})")
        if all_good:
            self.points += self.UNIT_TEST_COMPLETION_POINTS
            self.justification_parts.append(f"All tests passed: {self.UNIT_TEST_COMPLETION_POINTS}")
        else:
            gently("Failing instructor unit tests")
            self.justification_parts.append(f"Some tests failed: {test_points}")
        return all_good

